from unittest import TestCase

from musurgia.fractaltree.fractalmusic import FractalMusic
from musurgia.fractaltree.fractaltree import FractalTree


class Test(TestCase):
    def test_1(self):
        ft = FractalTree(value=100)
        self.assertEqual(100, ft.value)

    def test_2(self):
        fm = FractalMusic(value=100)
        self.assertEqual(100, fm.value)

    def test_3(self):
        fm = FractalMusic(duration=100)
        self.assertEqual(100, fm.value)

    def test_4(self):
        fm = FractalMusic(value=100)
        self.assertEqual(100, fm.duration)



from unittest import TestCase

from musurgia.random import Random


class Test(TestCase):
    def test_1(self):
        pool = [1, 3, 2, 4, 5, 1, 1]
        r = Random(pool=pool, periodicity=2, seed=20)
        output = [r.__next__() for i in range(20)]
        expected = [3, 2, 1, 5, 3, 1, 4, 3, 2, 4, 5, 3, 2, 4, 1, 5, 4, 1, 3, 5]
        self.assertEqual(expected, output)

    def test_2(self):
        pool = [1, 3, 2, 4, 5, 1, 1, 1, 1, 1, 1]
        r = Random(pool=pool, periodicity=0, seed=20)
        expected = [1, 3, 2, 4, 5]
        self.assertEqual(expected, r.pool)

    def test_3(self):
        pool = [1, 3, 2, 4, 5, 6]
        r = Random(pool=pool, periodicity=0, seed=20)
        output = [r.__next__() for i in range(20)]
        expected = [6, 6, 3, 2, 6, 6, 1, 2, 5, 3, 1, 4, 4, 1, 1, 3, 2, 4, 5, 4]
        self.assertEqual(expected, output)

    def test_4(self):
        pool = [1, 3, 2, 4, 5, 6]
        r = Random(pool=pool, periodicity=3, seed=30)
        output = [r.__next__() for i in range(20)]
        expected = [5, 2, 1, 6, 3, 2, 1, 4, 6, 3, 1, 4, 5, 3, 1, 4, 6, 2, 5, 1]
        self.assertEqual(expected, output)

    def test_5(self):
        pool = [1, 3, 2, 4, 5, 6]
        r1 = Random(pool=pool, periodicity=3, seed=30)
        r2 = Random(pool=pool, periodicity=5, seed=30)
        output = [r1.__next__() for i in range(20)]
        expected = [5, 2, 1, 6, 3, 2, 1, 4, 6, 3, 1, 4, 5, 3, 1, 4, 6, 2, 5, 1]
        self.assertEqual(expected, output)


import os

from musicscore.musictree.treescoretimewise import TreeScoreTimewise

from musurgia.unittest import TestCase
from musurgia.fractaltree.fractalmusic import FractalMusic

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.fm = FractalMusic(tempo=60, quarter_duration=4)

    def test_1(self):
        self.fm.split(1, 2)
        score = TreeScoreTimewise()
        self.fm.get_score(score)
        xml_path = path + '_test_1.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_2(self):
        self.fm.split(1, 2)[1].chord.to_rest()
        score = TreeScoreTimewise()
        self.fm.get_score(score)
        xml_path = path + '_test_2.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)



import os

from musurgia.unittest import TestCase
from musurgia.arithmeticprogression import ArithmeticProgression
from musurgia.fractaltree.fractalmusicsquare import Square

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.square = Square(duration=600, tree_permutation_order=[4, 1, 2, 3], proportions=[1, 2, 3, 4])
        # for module in self.square.modules.values():
        #     module.tempo = 72
        tempi = [round(tempo) for tempo in ArithmeticProgression(a1=40, an=72, n=4)]
        for row in self.square.rows:
            for module in row.modules:
                tempo = tempi[row.number - 1]
                module.tempo = tempo

    def test_1(self):
        text_path = path + '_test_1.txt'
        self.square.write_info(text_path=text_path, show_attributes=['tempo', 'quarter_duration', 'duration'])
        self.assertCompareFiles(text_path)

    def test_2(self):
        text_path = path + '_test_2.txt'
        module = self.square.get_module(1, 1)
        module.duration = 3

        self.square.write_info(text_path=text_path, show_attributes=['tempo', 'quarter_duration', 'duration'])
        self.assertCompareFiles(text_path)

    def test_3(self):
        self.square.change_module_quarter_duration(1, 1, 3)

        text_path = path + '_test_3.txt'

        self.square.write_info(text_path=text_path, show_attributes=['tempo', 'quarter_duration', 'duration'])

        self.assertCompareFiles(text_path)

    def test_4(self):
        self.square = Square(duration=600, tree_permutation_order=[3, 4, 2, 1], proportions=[1, 2, 3, 4],
                             reading_direction='vertical')
        self.square.change_module_duration(1, 1, 3)

        orders = [module.permutation_order for module in self.square.get_all_modules()]
        result = [[4, 1, 2, 3], [3, 2, 1, 4], [1, 3, 4, 2], [2, 4, 3, 1], [2, 3, 1, 4], [1, 4, 2, 3], [4, 2, 3, 1],
                  [3, 1, 4, 2], [1, 4, 3, 2], [2, 3, 4, 1], [3, 1, 2, 4], [4, 2, 1, 3], [3, 2, 4, 1], [4, 1, 3, 2],
                  [2, 4, 1, 3], [1, 3, 2, 4]]

        self.assertEqual(result, orders)

    def test_5(self):
        row_1 = self.square.rows[0]
        row_1.change_module_duration(1, 3)
        text_path = path + '_test_5.txt'
        self.square.write_info(text_path=text_path, show_attributes=['tempo', 'quarter_duration', 'duration'])
        self.assertCompareFiles(text_path)

    def test_6(self):
        row_2 = self.square.rows[1]
        row_2.change_module_quarter_duration(2, 3)
        text_path = path + '_test_6.txt'
        self.square.write_info(text_path=text_path, show_attributes=['tempo', 'quarter_duration', 'duration'])
        self.assertCompareFiles(text_path)

    def test_7(self):
        self.assertEqual(None, self.square.name)

    def test_8(self):
        self.square.name = 'blue'
        self.assertEqual('blue', self.square.name)

    def test_9(self):
        self.square.name = 'blue'
        text_path = path + '_test_9.txt'
        self.square.write_info(text_path=text_path, show_attributes=['duration'],
                               title='square: {}'.format(self.square.name))
        self.assertCompareFiles(text_path)

    def test_10(self):
        self.square.name = 'blue'
        row_2 = self.square.rows[1]
        row_2.change_module_quarter_duration(2, 3)
        copied = self.square.__deepcopy__()
        self.square.name = 'red'
        copied.name = 'green'
        text_path = path + '_test_10.txt'
        self.square.write_info(text_path=text_path, show_attributes=['tempo', 'quarter_duration', 'duration'],
                               title='square: {}'.format(self.square.name))
        self.assertCompareFiles(text_path)



from musurgia.pdf.font import Font
from musurgia.unittest import TestCase


class TestFont(TestCase):
    def setUp(self) -> None:
        self.font = Font()

    def test_load_afm(self):
        afm = self.font._afm
        actual = afm.get_familyname()
        expected = self.font.family
        self.assertEqual(expected, actual)

    def test_text_pixel_height(self):
        self.font.size = 12
        actual = self.font.get_text_pixel_height('This One')
        expected = 8.844
        self.assertEqual(expected, actual)

    def test_text_pixel_width(self):
        self.font.size = 11
        actual = self.font.get_text_pixel_width('This One')
        expected = 44.627
        self.assertEqual(expected, actual)



class PdfUnit:
    _DEFAULT_UNIT = 'mm'
    GLOBAL_UNIT = _DEFAULT_UNIT

    @staticmethod
    def get_k():
        k_dict = {'pt': 1, 'mm': 72 / 25.4, 'cm': 72 / 2.54, 'in': 72.}
        k = k_dict.get(PdfUnit.GLOBAL_UNIT)
        if k is None:
            raise AttributeError(f'wrong unit {PdfUnit.GLOBAL_UNIT}')
        return k

    @staticmethod
    def reset():
        PdfUnit.GLOBAL_UNIT = PdfUnit._DEFAULT_UNIT

    @staticmethod
    def change(val):
        PdfUnit.GLOBAL_UNIT = val
#
# def _get_unit(k):
#     unit_dict = {1: 'pt', 72. / 25.4: 'mm', 72. / 2.54: 'cm', 72.: 'in'}
#     unit = unit_dict.get(k)
#     if k is None:
#         raise AttributeError(f'wrong k {k}')
#     return unit



from pathlib import Path

from musurgia.pdf.line import VerticalLineSegment
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.rowcolumn import DrawObjectRow
from musurgia.unittest import TestCase

path = Path(__file__)
# class TestRow(TestCase):
#     def setUp(self) -> None:
#         self.pdf = Pdf()
#
#     def test_add_label_lef(self):
#         row = DrawObjectRow()
#         row.add_draw_object(VerticalLineSegment(10))


from musurgia.random import Random

ar_1 = Random(pool=[5, 2, 3, 1, 2, 1])
print(ar_1.pool)
ar_2 = Random(pool=[5.3, 2.2, 3, 1, 2, 1])
print(ar_2.pool)
ar_3 = Random(pool=['1', '2', '3', '4'])
print(ar_3.pool)
ar_4 = Random(pool=[4, 3, 5.3, 2.2, 3, 1, 2, 1])
print(ar_4.pool)
#
# pool = ['1', '2', '3', '4']
# print(sorted(list(set(pool))))




import itertools

from quicktions import Fraction

from musurgia.arithmeticprogression import ArithmeticProgression
from musurgia.basic_functions import flatten
from musurgia.pdf.line import HorizontalLineSegment, HorizontalSegmentedLine
from musurgia.pdf.newdrawobject import DrawObject
from musurgia.pdf.rowcolumn import DrawObjectColumn, DrawObjectRow
from musurgia.permutation import LimitedPermutation, permute
from musurgia.tree import Tree


class FractalTreeException(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class ValueCannotBeChanged(FractalTreeException):
    def __init__(self, *args):
        super().__init__(*args)


class SetValueFirst(FractalTreeException):
    def __init__(self, *args):
        super().__init__('FractalTree().value must be set before add_layer()', *args)


class _Graphic(DrawObject):
    def __init__(self, fractal_tree, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._draw_object_column = None
        self._fractal_tree = fractal_tree
        self._unit = 1
        self._large_mark_line_max_length = 6
        self._large_mark_line_min_length = 3

    def update_draw_object_columns(self):
        unit = self.unit
        max_large_ml = self.large_mark_line_max_length
        min_large_ml = self.large_mark_line_min_length

        def _make_segment():
            segment = HorizontalLineSegment(length=node.value * unit)
            ml_length = (max_large_ml - (
                    node.get_distance() * (max_large_ml - min_large_ml) / self._fractal_tree.number_of_layers)) / 2
            segment.start_mark_line.length = ml_length
            segment.end_mark_line.length = ml_length
            return segment

        for node in self._fractal_tree.traverse():
            node.graphic._draw_object_column = DrawObjectColumn()
            node.graphic._draw_object_column.add_draw_object(_make_segment())

            if node.get_children():
                node.graphic._draw_object_column.add_draw_object(DrawObjectRow(top_margin=3))
            if node.up:
                node.up.graphic._draw_object_column.draw_objects[1].add_draw_object(node.graphic)
            if not node.up or node.up.get_children().index(node) == len(node.up.get_children()) - 1:
                node.graphic._draw_object_column.draw_objects[0].end_mark_line.show = True
                node.graphic._draw_object_column.draw_objects[0].end_mark_line.length *= 2
            if not node.up or node.up.get_children().index(node) == 0:
                node.graphic._draw_object_column.draw_objects[0].start_mark_line.length *= 2

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, val):
        self._unit = val

    @property
    def large_mark_line_max_length(self):
        return self._large_mark_line_max_length

    @large_mark_line_max_length.setter
    def large_mark_line_max_length(self, val):
        self._large_mark_line_max_length = val

    @property
    def large_mark_line_min_length(self):
        return self._large_mark_line_min_length

    @large_mark_line_min_length.setter
    def large_mark_line_min_length(self, val):
        self._large_mark_line_min_length = val

    @property
    def draw_object_column(self):
        if self._draw_object_column is None:
            self.update_draw_object_columns()
        return self._draw_object_column

    def add_labels(self, function, **kwargs):
        for node in self._fractal_tree.traverse():
            node.graphic.get_straight_line().add_label(function(node), **kwargs)

    def get_start_mark_line(self):
        return self.draw_object_column.draw_objects[0].start_mark_line

    def get_end_mark_line(self):
        return self.draw_object_column.draw_objects[0].end_mark_line

    def get_straight_line(self):
        return self.draw_object_column.draw_objects[0].straight_line

    def get_relative_x2(self):
        return self.draw_object_column.get_relative_x2()

    def get_relative_y2(self):
        return self.draw_object_column.get_relative_y2()

    def get_line_segment(self):
        return self.draw_object_column.draw_objects[0]

    def get_children_draw_object_row(self):
        return self.draw_object_column.draw_objects[1]

    def change_segment_attributes(self, **kwargs):
        for node in self._fractal_tree.traverse():
            for key in kwargs:
                setattr(node.graphic.get_line_segment(), key, kwargs[key])

    def draw(self, pdf):
        self.draw_object_column.draw(pdf)


class FractalTree(Tree):
    def __init__(self, value=None, proportions=None, tree_permutation_order=None, multi=None, fertile=True,
                 reading_direction='horizontal', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._value = None
        self._proportions = None
        self._tree_permutation_order = None
        self._multi = None
        self._permutation_order = None
        self._fertile = None
        self._reading_direction = None
        self._fractal_order = None
        self._children_fractal_values = None
        self._name = None
        self._graphic = _Graphic(self)

        self.proportions = proportions
        self.tree_permutation_order = tree_permutation_order
        self.value = value
        self.multi = multi
        self.fertile = fertile
        self.reading_direction = reading_direction

    # // private methods

    def _calculate_children_fractal_orders(self):
        if self.value and self.proportions:
            children_fractal_orders = range(1, self.size + 1)
            permutation_order = self.permutation_order
            if permutation_order:
                children_fractal_orders = permute(children_fractal_orders, permutation_order)
            return children_fractal_orders

    def _calculate_children_fractal_values(self):
        value = self.value
        if value and self.proportions:
            children_fractal_values = [value * prop for prop in self.proportions]
            if self.permutation_order:
                try:
                    children_fractal_values = permute(children_fractal_values, self.permutation_order)
                except ValueError:
                    raise ValueError('proportions and tree_permutation_order should have the same length')
            return children_fractal_values

    def _calculate_position_in_tree(self):
        parent = self.up
        if self.is_root:
            return 0
        else:
            index = parent.get_children().index(self)
            if index == 0:
                return parent.position_in_tree
            else:
                return parent.get_children()[index - 1].position_in_tree + parent.get_children()[index - 1].value

    def _change_children_value(self, factor):
        for child in self.get_children():
            child._value *= factor
            child._change_children_value(factor)

    def _check_merge_nodes(self, nodes):
        return True

    def _child_multi(self, parent, index):
        number_of_children = self.size
        multi_first = sum(parent.multi) % number_of_children
        if multi_first == 0:
            multi_first = number_of_children
        ch_multi = (multi_first, index + 1)
        return ch_multi

    def _get_merge_lengths(self, number_of_children, merge_index):
        if 0 > merge_index > self.size - 1:
            raise ValueError('generate_children.merge_index {} not valid'.format(merge_index))
        if number_of_children > self.size or number_of_children < 0:
            raise ValueError(
                'generate_children.number_of_children {} must be a positive int'.format(number_of_children))
        if number_of_children == 1:
            return [self.size]

        lengths = self.size * [1]
        pointer = merge_index
        sliced_lengths = [lengths[:pointer], lengths[pointer:]]

        if not sliced_lengths[0]:
            sliced_lengths = sliced_lengths[1:]

        while len(sliced_lengths) < number_of_children and len(sliced_lengths[0]) > 1:
            temp = sliced_lengths[0]
            sliced_lengths[0] = temp[:-1]
            sliced_lengths.insert(1, temp[-1:])

        while len(sliced_lengths) < number_of_children and len(sliced_lengths[pointer]) > 1:
            temp = sliced_lengths[pointer]
            sliced_lengths[pointer] = temp[:-1]
            sliced_lengths.insert(pointer + 1, temp[-1:])

        sliced_lengths = [len(x) for x in sliced_lengths]

        return sliced_lengths

    # // properties
    @property
    def children_fractal_values(self):
        if not self._children_fractal_values:
            self._children_fractal_values = self._calculate_children_fractal_values()
        return self._children_fractal_values

    @property
    def children_fractal_orders(self):
        return self._calculate_children_fractal_orders()

    @property
    def fertile(self):
        return self._fertile

    @fertile.setter
    def fertile(self, val):
        if not isinstance(val, bool):
            raise TypeError('fertile.value must be of type bool not{}'.format(type(val)))
        self._fertile = val

    @property
    def fractal_order(self):
        return self._fractal_order

    @property
    def layers(self):
        return [self.get_layer(i) for i in range(self.number_of_layers + 1)]

    @property
    def multi(self):
        return self._multi

    @multi.setter
    def multi(self, vals):
        if vals is None:
            vals = (1, 1)

        elif not isinstance(vals, tuple) or len(vals) != 2:
            raise TypeError('multi has to be a tuple with a length of 2')

        m_1 = vals[0]
        m_2 = vals[1]
        m_1, m_2 = (((m_1 - 1) % self.size) + 1 + ((m_2 - 1) // self.size)) % self.size, (
                (m_2 - 1) % self.size) + 1
        if m_1 == 0:
            m_1 = self.size
        self._multi = (m_1, m_2)

        self._permutation_order = None
        self._children_fractal_values = None

    @property
    def name(self):
        if self._name is None:
            if self.is_root:
                self._name = '0'
            elif self.up.is_root:
                self._name = str(self.up.get_children().index(self) + 1)
            else:
                self._name = self.up.name + '.' + str(self.up.get_children().index(self) + 1)
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def next(self):
        if not self.is_leaf():
            raise Exception('FractalTree().next property can only be used for leaves')
        else:
            try:

                parent = self.up
                while parent is not None:
                    parent_leaves = list(parent.traverse_leaves())
                    index = parent_leaves.index(self)
                    try:
                        return parent_leaves[index + 1]
                    except IndexError:
                        parent = parent.up
                raise IndexError()
            except IndexError:
                return None

    @property
    def number_of_layers(self):
        if self.get_leaves() == [self]:
            return 0
        else:
            return self.get_farthest_leaf().get_distance(self)

    @property
    def graphic(self):
        return self._graphic

    @property
    def permutation_order(self):
        def _calculate_permutation_order():
            if self.tree_permutation_order:
                permutation = LimitedPermutation(input_list=list(range(1, self.size + 1)),
                                                 main_permutation_order=self.tree_permutation_order, multi=self.multi,
                                                 reading_direction=self.reading_direction)

                return permutation.permutation_order
            else:
                raise FractalTreeException('tree_permutation_order must be set')

        if self._permutation_order is None:
            self._permutation_order = _calculate_permutation_order()
        return self._permutation_order

    @property
    def position(self):
        if self.is_root:
            # return self.first_position
            return 0
        else:
            siblings = self.up.get_children()
            index = siblings.index(self)
            previous_siblings = siblings[:index]
            position_in_parent = 0
            for child in previous_siblings: position_in_parent += child.value
            return position_in_parent + self.up.position

    @property
    def position_in_tree(self):
        return self._calculate_position_in_tree()

    @property
    def proportions(self):
        return self._proportions

    @proportions.setter
    def proportions(self, values):
        if values is None:
            values = [1, 2, 3]
        self._proportions = [Fraction(Fraction(value) / Fraction(sum(values))) for value in values]

    @property
    def reading_direction(self):
        return self._reading_direction

    @reading_direction.setter
    def reading_direction(self, val):
        if self._reading_direction:
            raise FractalTreeException('reading_direction can only be set during initialisation')
        permitted = ['horizontal', 'vertical']
        if val not in permitted:
            raise ValueError('reading_direction.value {} must be in {}'.format(val, permitted))
        self._reading_direction = val

    @property
    def size(self):
        return len(self.proportions)

    @property
    def tree_permutation_order(self):
        return self._tree_permutation_order

    @tree_permutation_order.setter
    def tree_permutation_order(self, values):
        if values is None:
            values = [3, 1, 2]
        self._tree_permutation_order = values
        self._permutation_order = None

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        if val:
            val = Fraction(val)
        if self._value:
            if self.is_root and not self.get_children():
                self._value = val
            else:
                raise ValueCannotBeChanged(
                    'FractalTree().value can only be changed for the root with no children. Use change_value() for other cases.'
                )
        else:
            self._value = val

        if self._value == 0:
            self.fertile = False

    # //public methods
    # add
    def add_layer(self, *conditions):
        if self.value is None:
            raise SetValueFirst()

        leaves = list(self.traverse_leaves())
        if not leaves:
            leaves = [self]

        if conditions:
            for leaf in leaves:
                for condition in conditions:
                    if not condition(leaf):
                        leaf.fertile = False
                        break

        for leaf in leaves:
            if leaf.fertile is True:
                for i in range(leaf.size):
                    new_node = leaf.copy()
                    new_node.value = leaf.children_fractal_values[i]
                    new_node.multi = self._child_multi(leaf, i)
                    leaf.add_child(new_node)
                    new_node._fractal_order = leaf.children_fractal_orders[i]

            else:
                pass

    def add_self(self):
        leaves = self.get_leaves()
        for leaf in leaves:
            new_node = self.copy()
            new_node._up = self
            new_node._fractal_order = self.fractal_order
            leaf.add_child(new_node)

    # get
    def get_layer(self, layer, key=None):
        if layer <= self.get_root().number_of_layers:
            branch_distances = []
            for child in self.get_children():
                branch_distances.append(child.get_farthest_leaf().get_distance())
            if layer == 0:
                if key:
                    if isinstance(key, str):
                        return [getattr(self, key)]
                    elif callable(key):
                        return [key(self)]
                    else:
                        raise TypeError(f'key: {key} must be either string or a callable object')
                else:
                    return [self]

            if layer >= 1:
                output = []
                if not branch_distances:
                    output.extend(self.get_layer(layer=layer - 1, key=key))
                for i in range(len(self.get_children())):
                    child = self.get_children()[i]
                    if branch_distances[i] == 1:
                        if key:
                            if isinstance(key, str):
                                output.append(getattr(child, key))
                            elif callable(key):
                                output.append(key(child))
                            else:
                                raise TypeError(f'key: {key} must be either string or a callable object')
                        else:
                            output.append(child)
                    else:
                        if layer == 1:
                            output.extend(child.get_layer(layer - 1, key))
                        else:
                            output.append(child.get_layer(layer - 1, key))

                return output
        else:
            err = 'max layer number=' + str(self.number_of_layers)
            raise ValueError(err)

    def get_next_sibling(self):
        try:
            siblings = self.up.get_children()
            index = siblings.index(self)
            return siblings[index + 1]
        except (IndexError, AttributeError):
            return None

    def get_previous_leaf(self):
        if not self.is_leaf():
            raise Exception('FractalTree().get_previous_leaf  can only be applied to leaves')
        try:
            parent = self.up
            while parent is not None:
                parent_leaves = list(parent.traverse_leaves())
                index = parent_leaves.index(self)
                try:
                    return parent_leaves[index - 1]
                except IndexError:
                    parent = parent.up
            raise IndexError()
        except IndexError:
            return None

    def get_previous_sibling(self):
        try:
            siblings = self.up.get_children()
            index = siblings.index(self)
            if index == 0:
                return None
            return siblings[index - 1]
        except AttributeError:
            return None

    # other

    def change_value(self, new_value):
        if not new_value:
            new_value = Fraction(new_value)
        if not self.value:
            self.value = new_value
        else:
            factor = Fraction(new_value, self.value)
            self._value = new_value
            for node in reversed(self.get_branch()[:-1]):
                node._value = sum([child.value for child in node.get_children()])

            self._change_children_value(factor)

    def generate_children(self, number_of_children, mode='reduce', merge_index=0):

        permitted_modes = ['reduce', 'reduce_backwards', 'reduce_forwards', 'reduce_sieve', 'merge']
        if mode not in permitted_modes:
            raise ValueError('generate_children.mode {} must be in{}'.format(mode, permitted_modes))
        if isinstance(number_of_children, int):
            if number_of_children > self.size:
                raise ValueError(
                    'generate_children.number_of_children {} can not be a greater than size {}'.format(
                        number_of_children, self.size))
            if number_of_children < 0:
                raise ValueError(
                    'generate_children.number_of_children {} must be a positive int'.format(number_of_children))
            elif number_of_children == 0:
                pass
            else:
                self.add_layer()
                if mode in ['reduce', 'reduce_backwards']:
                    self.reduce_children(
                        lambda child: child.fractal_order < self.size - number_of_children + 1)
                elif mode == 'reduce_forwards':
                    self.reduce_children(
                        lambda child: child.fractal_order > number_of_children)
                elif mode == 'reduce_sieve':
                    if number_of_children == 1:
                        self.reduce_children(condition=lambda child: child.fractal_order not in [1])
                    else:
                        ap = ArithmeticProgression(a1=1, an=self.size, n=number_of_children)
                        selection = [int(round(x)) for x in ap]
                        self.reduce_children(condition=lambda child: child.fractal_order not in selection)
                else:
                    merge_lengths = self._get_merge_lengths(number_of_children, merge_index)
                    self.merge_children(*merge_lengths)

        elif isinstance(number_of_children, tuple):
            self.generate_children(len(number_of_children), mode=mode, merge_index=merge_index)

            for index, child in enumerate(self.get_children()):
                if mode == 'reduce':
                    number_of_grand_children = number_of_children[
                        child.fractal_order - child.size + len(number_of_children) - 1]
                else:
                    number_of_grand_children = number_of_children[index]
                child.generate_children(number_of_grand_children, mode=mode, merge_index=merge_index)

        else:
            raise TypeError()

    def generate_layer_segmented_line(self, layer_number, unit, max_mark_line=6, shrink_factor=0.7):
        def get_segmented_line(layer_number):
            if layer_number > self.number_of_layers:
                layer_number = self.number_of_layers
            layer_nodes = flatten(self.get_layer(layer_number))
            lengths = [node.value * unit for node in layer_nodes]
            hsl = HorizontalSegmentedLine(lengths)
            for segment, node in zip(hsl.segments, layer_nodes):
                segment.node = node
                segment.start_mark_line.length = get_ml_length(node)
            return hsl

        def get_ml_length(node):
            if not node.up:
                return max_mark_line
            else:
                if node.up.get_children().index(node) == 0:
                    return get_ml_length(node.up)
                else:
                    return max_mark_line * shrink_factor / node.get_distance()

        def set_last_mark_line_length(row):
            last_mark_line = row.draw_objects[-1].end_mark_line
            last_mark_line.length = max_mark_line
            last_mark_line.show = True

        hls = get_segmented_line(layer_number)
        set_last_mark_line_length(hls)
        return hls

    def merge_children(self, *lengths):
        children = self.get_children()
        if not children:
            raise Exception('there are no children to be merged')
        if sum(lengths) != len(children):
            raise ValueError(
                'sum of lengths {} must be the same as length of children {}'.format(sum(lengths), len(children)))

        def _merge(nodes):
            self._check_merge_nodes(nodes)

            node_values = [node.value for node in nodes]

            nodes[0]._value = sum(node_values)
            for node in nodes[1:]:
                self.remove_child(node)

        iter_children = iter(children)
        chunks = [list(itertools.islice(iter_children, l)) for l in lengths]

        for chunk in chunks:
            _merge(chunk)

    def reduce_children(self, condition):
        if not self.get_children():
            raise ValueError(f'{self} has no children to be reduced')
        # for child in self.get_children():
        #     if condition(child):
        #         child._reduce()
        to_be_detached = [child for child in self.get_children() if condition(child)]
        for child in to_be_detached:
            child.detach()
        reduced_value = sum([child.value for child in self.get_children()])
        factor = self.value / reduced_value
        for child in self.get_children():
            new_value = child.value * factor
            child.change_value(new_value)

        self._children_fractal_values = [child.value for child in self.get_children()]

    # def split(self, *proportions):
    #
    #     for prop in proportions:
    #         self.add_self()
    #         new_value = self.value * prop / sum(proportions)
    #         self.get_children()[-1].change_value(new_value)

    def split(self, *proportions):
        if hasattr(proportions[0], '__iter__'):
            proportions = proportions[0]

        proportions = [Fraction(prop) for prop in proportions]

        for prop in proportions:
            value = self.value * prop / sum(proportions)
            new_node = self.copy()
            new_node.multi = self.multi
            new_node.change_value(value)
            new_node._fractal_order = self.fractal_order
            self.add_child(new_node)

        return self.get_children()

    # // copy

    def copy(self):
        return self.__class__(value=self.value, proportions=self.proportions,
                              tree_permutation_order=self.tree_permutation_order, multi=self.multi,
                              reading_direction=self.reading_direction, fertile=self.fertile)

    # def copy_without_children(self):

    def __deepcopy__(self, memodict={}):
        copied = self.__class__(value=self.value, proportions=self.proportions,
                                tree_permutation_order=self.tree_permutation_order, multi=self.multi,
                                reading_direction=self.reading_direction, fertile=self.fertile)
        copied._fractal_order = self.fractal_order
        copied._name = self._name
        for child in self.get_children():
            copied.add_child(child.__deepcopy__())
        return copied



from unittest import TestCase

from musurgia.permutation import LimitedPermutation, self_permute, get_self_multiplied_permutation, \
    get_reordered_self_multiplied_permutation, get_vertical_self_multiplied_permutation


class Test(TestCase):
    def test_1_1(self):
        permutation_order = [3, 1, 2]
        self_permutation = self_permute(permutation_order)
        result = [[3, 1, 2], [2, 3, 1], [1, 2, 3]]
        self.assertEqual(result, self_permutation)

    def test_1_2(self):
        permutation_order = [3, 1, 2]
        self_multiplied_permutation = get_self_multiplied_permutation(permutation_order)
        result = [[[1, 2, 3], [3, 1, 2], [2, 3, 1]],
                  [[2, 3, 1], [1, 2, 3], [3, 1, 2]],
                  [[3, 1, 2], [2, 3, 1], [1, 2, 3]]]
        self.assertEqual(result, self_multiplied_permutation)

    def test_1_3(self):
        permutation_order = [3, 1, 2]
        reordered_self_multiplied_permutation = get_reordered_self_multiplied_permutation(permutation_order)
        result = [[[3, 1, 2], [2, 3, 1], [1, 2, 3]],
                  [[1, 2, 3], [3, 1, 2], [2, 3, 1]],
                  [[2, 3, 1], [1, 2, 3], [3, 1, 2]]]
        self.assertEqual(result, reordered_self_multiplied_permutation)

    def test_1_4(self):
        permutation_order = [3, 1, 2]
        vertical_self_multiplied_permutation = get_vertical_self_multiplied_permutation(permutation_order)
        result = [[[1, 3, 2], [2, 1, 3], [3, 2, 1]],
                  [[2, 1, 3], [3, 2, 1], [1, 3, 2]],
                  [[3, 2, 1], [1, 3, 2], [2, 1, 3]]]
        self.assertEqual(result, vertical_self_multiplied_permutation)

    def test_2_1(self):
        perm = LimitedPermutation(['a', 'b', 'c'], [3, 1, 2], multi=[1, 1])
        result = [[3, 1, 2], [2, 3, 1], [1, 2, 3], [1, 2, 3], [3, 1, 2], [2, 3, 1], [2, 3, 1], [1, 2, 3], [3, 1, 2]]
        self.assertEqual(result, perm.multiplied_orders)

    def test_2_2(self):
        perm = LimitedPermutation(['a', 'b', 'c'], [3, 1, 2], multi=[1, 1], reading_direction='vertical')

        result = [[1, 3, 2], [2, 1, 3], [3, 2, 1], [2, 1, 3], [3, 2, 1], [1, 3, 2], [3, 2, 1], [1, 3, 2], [2, 1, 3]]
        self.assertEqual(result, perm.multiplied_orders)

    def test_3_1(self):
        permutation_order = [3, 4, 2, 1]
        self_multiplied_permutation = get_self_multiplied_permutation(permutation_order)
        result = [[[4, 3, 1, 2], [1, 2, 3, 4], [2, 1, 4, 3], [3, 4, 2, 1]],
                  [[2, 1, 4, 3], [3, 4, 2, 1], [1, 2, 3, 4], [4, 3, 1, 2]],
                  [[1, 2, 3, 4], [4, 3, 1, 2], [3, 4, 2, 1], [2, 1, 4, 3]],
                  [[3, 4, 2, 1], [2, 1, 4, 3], [4, 3, 1, 2], [1, 2, 3, 4]]]
        self.assertEqual(result, self_multiplied_permutation)

    def test_3_2(self):
        permutation_order = [3, 4, 2, 1]
        reordered_self_multiplied_permutation = get_reordered_self_multiplied_permutation(permutation_order)
        result = [[[3, 4, 2, 1], [2, 1, 4, 3], [4, 3, 1, 2], [1, 2, 3, 4]],
                  [[4, 3, 1, 2], [1, 2, 3, 4], [2, 1, 4, 3], [3, 4, 2, 1]],
                  [[2, 1, 4, 3], [3, 4, 2, 1], [1, 2, 3, 4], [4, 3, 1, 2]],
                  [[1, 2, 3, 4], [4, 3, 1, 2], [3, 4, 2, 1], [2, 1, 4, 3]]]
        self.assertEqual(result, reordered_self_multiplied_permutation)

    def test_3_3(self):
        permutation_order = [3, 4, 2, 1]
        vertical_self_multiplied_permutation = get_vertical_self_multiplied_permutation(permutation_order)
        result = [[[4, 1, 2, 3], [3, 2, 1, 4], [1, 3, 4, 2], [2, 4, 3, 1]],
                  [[2, 3, 1, 4], [1, 4, 2, 3], [4, 2, 3, 1], [3, 1, 4, 2]],
                  [[1, 4, 3, 2], [2, 3, 4, 1], [3, 1, 2, 4], [4, 2, 1, 3]],
                  [[3, 2, 4, 1], [4, 1, 3, 2], [2, 4, 1, 3], [1, 3, 2, 4]]]
        self.assertEqual(result, vertical_self_multiplied_permutation)

    def test_4_1(self):
        permutation_order = [4, 3, 2, 1]
        self_multiplied_permutation = get_self_multiplied_permutation(permutation_order)
        result = [[[1, 2, 3, 4], [4, 3, 2, 1], [1, 2, 3, 4], [4, 3, 2, 1]],
                  [[4, 3, 2, 1], [1, 2, 3, 4], [4, 3, 2, 1], [1, 2, 3, 4]],
                  [[1, 2, 3, 4], [4, 3, 2, 1], [1, 2, 3, 4], [4, 3, 2, 1]],
                  [[4, 3, 2, 1], [1, 2, 3, 4], [4, 3, 2, 1], [1, 2, 3, 4]]]
        self.assertEqual(result, self_multiplied_permutation)

    def test_4_2(self):
        permutation_order = [4, 3, 2, 1]
        vertical_self_multiplied_permutation = get_vertical_self_multiplied_permutation(permutation_order)
        result = [[[1, 4, 1, 4], [2, 3, 2, 3], [3, 2, 3, 2], [4, 1, 4, 1]],
                  [[4, 1, 4, 1], [3, 2, 3, 2], [2, 3, 2, 3], [1, 4, 1, 4]],
                  [[1, 4, 1, 4], [2, 3, 2, 3], [3, 2, 3, 2], [4, 1, 4, 1]],
                  [[4, 1, 4, 1], [3, 2, 3, 2], [2, 3, 2, 3], [1, 4, 1, 4]]]
        self.assertEqual(result, vertical_self_multiplied_permutation)

    def test_5(self):
        size = 3
        tree_permutation_order = (3, 1, 2)
        multi = (3, 4)

        permutation = LimitedPermutation(input_list=list(range(1, size + 1)),
                                         main_permutation_order=tree_permutation_order,
                                         multi=multi)
        result = permutation.multi
        expected = (1, 1)
        self.assertEqual(expected, result)

    def test_6(self):
        size = 3
        tree_permutation_order = (3, 1, 2)
        multi = (2, 4)

        permutation = LimitedPermutation(input_list=list(range(1, size + 1)),
                                         main_permutation_order=tree_permutation_order,
                                         multi=multi)
        result = permutation.multi
        expected = (3, 1)
        self.assertEqual(expected, result)



class PositionedObjectError(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class RelativePositionNotSettableError(PositionedObjectError):
    pass


class RelativeXNotSettableError(RelativePositionNotSettableError):
    pass


class RelativeYNotSettableError(RelativePositionNotSettableError):
    pass


class Positioned(object):
    def __init__(self, relative_x=None, relative_y=None, *args, **kwargs):
        self._relative_x = None
        self._relative_y = None

        self.relative_x = relative_x
        self.relative_y = relative_y
        super().__init__(*args, **kwargs)

    @property
    def relative_x(self):
        if self._relative_x is None:
            return 0
        return self._relative_x

    @relative_x.setter
    def relative_x(self, val):
        self._relative_x = val

    @property
    def relative_y(self):
        if self._relative_y is None:
            return 0
        return self._relative_y

    @relative_y.setter
    def relative_y(self, val):
        self._relative_y = val



from musurgia.fractaltree.fractalmusic import FractalMusic

fm = FractalMusic(quarter_duration=10, tempo=60)
fm.midi_generator.midi_range = [60, 84]
fm.add_layer()
node = fm.get_children()[0]

copy = node.__deepcopy__()
print(copy.midi_generator.midi_range)
# copy.midi_generator.midi_range = node.midi_generator.midi_range
node.add_layer()
print([child.midi_value for child in node.get_children()])

copy.add_layer()
print([child.midi_value for child in copy.get_children()])



import os

from prettytable import PrettyTable


class Table(PrettyTable):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def writ_to_path(self, table_path, title=None):
        os.system('touch ' + table_path)
        file = open(table_path, 'w')
        if title:
            file.write(title + '\n')

        file.write(self.get_string())
        file.close()

    def extend(self, other_table):
        if not isinstance(other_table, Table):
            raise TypeError()
        for row in other_table._rows:
            self.add_row(row)

    def get_row(self, index):
        return self._rows[index]

    @property
    def rows(self):
        return self._rows



from unittest import TestCase

from musurgia.fractaltree.fractaltree import FractalTree


class Test(TestCase):
    def setUp(self) -> None:
        ft = FractalTree(value=10, reading_direction='vertical')
        ft.add_layer()
        ft.add_layer()
        self.ft = ft.get_children()[1]
        self.deep_copied = self.ft.__deepcopy__()

    def test(self, exp=None, act=None):
        if not exp:
            exp = self.ft

        if not act:
            act = self.deep_copied

        self.assertEqual(exp.value, act.value)
        self.assertEqual(exp.proportions, act.proportions)
        self.assertEqual(exp.value, act.value)
        self.assertEqual(exp.proportions, act.proportions)
        self.assertEqual(exp.tree_permutation_order, act.tree_permutation_order)
        self.assertEqual(exp.fractal_order, act.fractal_order)
        self.assertEqual(exp.reading_direction, act.reading_direction)
        self.assertEqual(exp._name, act._name)

    def test_1(self):
        self.assertIsNone(self.deep_copied.up)



from itertools import cycle

from musicscore.musicstream.streamvoice import SimpleFormat

from musurgia.unittest import TestCase
from musurgia.arithmeticprogression import ArithmeticProgression
from musurgia.chordfield.valuegenerator import ValueGenerator, NoDurationError, PositionError
from musurgia.interpolation import RandomInterpolation


class Test(TestCase):
    def test_1(self):
        vg = ValueGenerator(generator=ArithmeticProgression(a1=0.2, an=1, s=20, correct_s=True), duration=100)
        expected = 20
        actual = sum(list(vg))
        self.assertEqual(expected, actual)

    def test_2(self):
        with self.assertRaises(NoDurationError):
            vg = ValueGenerator(generator=ArithmeticProgression(a1=0.2, an=1, s=10, correct_s=True))
            vg.__next__()

    def test_3(self):
        vg = ValueGenerator(generator=ArithmeticProgression(a1=0.2, an=1, s=10, correct_s=True), duration=20,
                            value_mode='duration')
        expected = 20
        actual = sum(list(vg))
        self.assertEqual(expected, actual)

    def test_4(self):
        vg = ValueGenerator(generator=cycle([1]), duration=20)
        vg.position = 20
        with self.assertRaises(StopIteration):
            vg.__next__()

    def test_5(self):
        vg = ValueGenerator(generator=cycle([1]), duration=20)
        expected = 1
        actual = vg(19)
        self.assertEqual(expected, actual)

    def test_6(self):
        vg = ValueGenerator(generator=cycle([1]), duration=20)
        with self.assertRaises(PositionError):
            vg(20)

    def test_7(self):
        vg = ValueGenerator(generator=cycle([1]), duration=20)
        with self.assertRaises(PositionError):
            vg(20)

    # def test_8(self):
    #     vg = ValueGenerator(generator=ArithmeticProgression(a1=0.2, an=1, s=10), duration=20)
    #     with self.assertRaises(CallConflict):
    #         vg(21)

    def test_9(self):
        vg = ValueGenerator(generator=RandomInterpolation(start=[1, 1, 2], end=[3, 4, 5], seed=10),
                            duration=10)
        actual = [vg(x / 2) for x in range(20)]
        expected = [1, 2, 2, 1, 1, 2, 3, 1, 3, 2, 3, 2, 4, 3, 2, 3, 2, 5, 4, 5]
        self.assertEqual(expected, actual)

    # def test_10(self):
    #     vg = ValueGenerator(generator=RandomInterpolation(start=[1, 1, 2], end=[3, 4, 5], seed=10),
    #                         duration=10)
    #     with self.assertRaises(GeneratorHasNoNextError):
    #         vg.__next__()

    # def test_11(self):
    #     vg = ValueGenerator(generator=RandomInterpolation(start=[1, 1, 2], end=[3, 4, 5], seed=10),
    #                         duration=10)
    #     with self.assertRaises(GeneratorHasNoNextError):
    #         list(vg)

    def test_12(self):
        vg = ValueGenerator(generator=RandomInterpolation(start=[1, 1, 2], end=[3, 4, 5], seed=10),
                            value_mode='duration',
                            duration=10)

        expected = [1, 2, 2, 4, 3]
        actual = list(vg)
        self.assertEqual(expected, actual)

    def test_13(self):
        vg = ValueGenerator(generator=iter([1, 2, 3, 2, 1, 2, 1, 2, 2]),
                            value_mode='duration', duration=10)
        expected = [1, 2, 3, 2, 1, 2]
        actual = list(vg)
        self.assertEqual(expected, actual)

    def test_14(self):
        vg = ValueGenerator(generator=iter(SimpleFormat(quarter_durations=[1, 2, 3, 2, 1, 2, 1, 2, 2]).chords),
                            value_mode='chord', duration=10)

        expected = [1, 2, 3, 2, 1, 2]
        actual = [ch.quarter_duration for ch in list(vg)]
        self.assertEqual(expected, actual)

    def test_15(self):
        vgg = ValueGenerator()
        vg_1 = ValueGenerator(generator=cycle([2]), duration=5)
        vg_2 = ValueGenerator(generator=cycle([3]), duration=10)
        vgg.add_child(vg_1)
        vgg.add_child(vg_2)
        actual = vgg.__next__()
        expected = 2
        self.assertEqual(expected, actual)

    def test_16(self):
        vgg = ValueGenerator()
        vg_1 = ValueGenerator(generator=cycle([2]), duration=5)
        vg_2 = ValueGenerator(generator=cycle([3]), duration=10)
        vgg.add_child(vg_1)
        vgg.add_child(vg_2)
        actual = vgg(6)
        expected = 3
        self.assertEqual(expected, actual)

    def test_17(self):
        vgg = ValueGenerator()
        vg_1 = ValueGenerator(generator=ArithmeticProgression(a1=0.2, an=1, correct_s=True), duration=10,
                              value_mode='duration')
        vg_2 = ValueGenerator(generator=ArithmeticProgression(an=0.2, a1=1, correct_s=True), duration=5,
                              value_mode='duration')
        vgg.add_child(vg_1)
        vgg.add_child(vg_2)
        actual = float(vgg(12))
        expected = 1.0416666666666667
        self.assertEqual(expected, actual)

    def test_18(self):
        vgg = ValueGenerator()
        vg_1 = ValueGenerator(generator=ArithmeticProgression(a1=0.2, an=1, correct_s=True), duration=10,
                              value_mode='duration')
        vg_2 = ValueGenerator(generator=ArithmeticProgression(an=0.2, a1=1, correct_s=True), duration=5,
                              value_mode='duration')
        vgg.add_child(vg_1)
        vgg.add_child(vg_2)
        vgg(12)
        actual = float(vgg.position)
        expected = 13.041666666666666
        self.assertEqual(expected, actual)

    def test_19(self):
        vgg = ValueGenerator()
        vg_1 = ValueGenerator(generator=ArithmeticProgression(a1=0.2, an=1, correct_s=True), duration=10,
                              value_mode='duration')
        vg_2 = ValueGenerator(generator=ArithmeticProgression(an=0.2, a1=1, correct_s=True), duration=5,
                              value_mode='duration')
        vgg.add_child(vg_1)
        vgg.add_child(vg_2)
        actual = [round(float(x), 3) for x in vgg]
        expected = [0.208, 0.264, 0.319, 0.375, 0.431, 0.486, 0.542, 0.597, 0.653, 0.708, 0.764, 0.819, 0.875, 0.931,
                    0.986, 1.042, 1.042, 0.923, 0.804, 0.685, 0.565, 0.446, 0.327, 0.208]
        self.assertEqual(expected, actual)

    def test_20(self):
        vgg = ValueGenerator()
        vg_1 = ValueGenerator(generator=cycle([2]), duration=5)
        vg_2 = ValueGenerator(generator=cycle([3]), duration=10)
        vgg.add_child(vg_1)
        vgg.add_child(vg_2)
        actual = vgg.duration
        expected = 15
        self.assertEqual(expected, actual)

    def test_21(self):
        vgg = ValueGenerator()
        vg_1 = ValueGenerator(generator=cycle([2]), duration=5)
        vg_2 = ValueGenerator(generator=cycle([3]), duration=10)
        vgg.add_child(vg_1)
        vgg.add_child(vg_2)

        copied = vgg.__deepcopy__()
        actual = copied.duration
        expected = 15
        self.assertEqual(expected, actual)



import os
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

from musurgia.fractaltree.fractalmusic import FractalMusic, MergeTempoException
from musurgia.unittest import TestCase

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.fm = FractalMusic(proportions=(1, 2, 3, 4, 5), tree_permutation_order=(3, 5, 1, 2, 4))
        self.fm.duration = 10
        self.fm.midi_generator.midi_range = [60, 72]

    def test_1(self):
        fm = self.fm
        fm.add_layer()
        fm.tempo = 60
        for ch in fm.get_children():
            ch.chord.add_words(ch.fractal_order)

        score = TreeScoreTimewise()
        fm.get_simple_format().to_stream_voice().add_to_score(score)

        fm.merge_children(2, 1, 2)
        # print(fm.get_leaves(key=lambda leaf: leaf.index))
        # print(fm.get_leaves(key=lambda leaf: leaf.fractal_order))
        # print(fm.get_leaves(key=lambda leaf: round(float(leaf.value), 2)))
        # print(fm.get_leaves(key=lambda leaf: round(float(leaf.duration), 2)))
        # print(fm.get_leaves(key=lambda leaf: round(float(leaf.chord.quarter_duration), 2)))
        fm.get_simple_format().to_stream_voice().add_to_score(score, part_number=2)
        xml_path = path + '_test_1.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)
        # # ft.add_layer()
        # # print(ft.get_leaves(key=lambda leaf: leaf.index))
        # # print(ft.get_leaves(key=lambda leaf: leaf.fractal_order))
        # # print(ft.get_leaves(key=lambda leaf: round(float(leaf.value), 2)))
        # ft.merge_children(1, 2, 2)
        # # print(ft.get_leaves(key=lambda leaf: leaf.index))
        # self.assertEqual(ft.get_leaves(key=lambda leaf: leaf.fractal_order), [3, 5, 2])
        # self.assertEqual(ft.get_leaves(key=lambda leaf: round(float(leaf.value), 2)), [2.0, 4.0, 4.0])

    def test_2(self):
        self.fm.duration = 20
        self.fm.add_layer()
        self.fm.get_children()[1].tempo = 72
        self.fm.set_none_tempi(60)
        with self.assertRaises(MergeTempoException):
            self.fm.merge_children(2, 1, 2)



from unittest import TestCase

from musurgia.fractaltree.fractaltree import FractalTree


class Test(TestCase):
    def setUp(self) -> None:
        self.ft = FractalTree(value=10)

    def test_1(self):
        # with self.assertRaises(ValueError):
        self.ft.generate_children(number_of_children=0)
        result = [None]
        self.assertEqual(result, self.ft.get_leaves(key=lambda leaf: leaf.fractal_order))

    def test_2(self):
        ft = self.ft
        ft.generate_children(number_of_children=1)
        result = [3]
        self.assertEqual(result, self.ft.get_leaves(key=lambda leaf: leaf.fractal_order))

    def test_3(self):
        ft = self.ft
        ft.generate_children(number_of_children=2)
        result = [3, 2]
        self.assertEqual(result, self.ft.get_leaves(key=lambda leaf: leaf.fractal_order))

    def test_4(self):
        ft = self.ft
        ft.generate_children(number_of_children=3)
        result = [3, 1, 2]
        self.assertEqual(result, self.ft.get_leaves(key=lambda leaf: leaf.fractal_order))

    def test_5(self):
        ft = self.ft
        with self.assertRaises(ValueError):
            ft.generate_children(number_of_children=4)

    def test_6(self):
        ft = self.ft
        ft.generate_children(number_of_children=(1, 1, 1))
        result = [[3], [3], [3]]
        self.assertEqual(result, self.ft.get_leaves(key=lambda leaf: leaf.fractal_order))

    def test_7(self):
        ft = self.ft
        ft.generate_children(number_of_children=(0, 1, 2))
        result = [[2, 3], 1, [3]]
        self.assertEqual(result, self.ft.get_leaves(key=lambda leaf: leaf.fractal_order))

    def test_8(self):
        ft = self.ft
        ft.generate_children(number_of_children=(1, 2, (1, 2, 3)))
        result = [
            [[3], [2, 3], [3, 1, 2]],
            [3],
            [2, 3]
        ]
        self.assertEqual(result, self.ft.get_leaves(key=lambda leaf: leaf.fractal_order))

    def test_9(self):
        ft = self.ft
        ft.generate_children(
            number_of_children=(
                (1, 3),
                2,
                (1,
                 (1, 3),
                 3)
            )
        )
        result = [
            [
                [3],
                [
                    [3],
                    [2, 3, 1]
                ],
                [3, 1, 2]
            ],
            [
                [3, 1, 2],
                [3]
            ],
            [2, 3]
        ]
        self.assertEqual(result, self.ft.get_leaves(key=lambda leaf: leaf.fractal_order))

    def test_10(self):
        ft = FractalTree(value=10, proportions=(1, 2, 3, 4, 5, 6, 7), tree_permutation_order=(2, 6, 4, 1, 3, 7, 5))
        ft.generate_children(mode='merge', number_of_children=2)
        test_case = ft.get_leaves(key=lambda leaf: float(leaf.value))
        expected = [8.214285714285714, 1.7857142857142858]
        self.assertEqual(expected, test_case)

    def test_11(self):
        ft = FractalTree(value=10, proportions=(1, 2, 3, 4, 5, 6, 7), tree_permutation_order=(2, 6, 4, 1, 3, 7, 5))
        output = []
        for i in range(7):
            ft_copy = ft.__deepcopy__()
            ft_copy.generate_children(mode='merge', number_of_children=2, merge_index=i)
            output.append(ft_copy.get_leaves(key=lambda leaf: float(leaf.value)))
        test_case = output
        # ft.add_layer()
        # print(ft.get_leaves(key=lambda leaf: float(leaf.value)))
        # print(output)
        expected = [[8.214285714285714, 1.7857142857142858],
                    [0.7142857142857143, 9.285714285714286],
                    [2.857142857142857, 7.142857142857143],
                    [4.285714285714286, 5.714285714285714],
                    [4.642857142857143, 5.357142857142857],
                    [5.714285714285714, 4.285714285714286],
                    [8.214285714285714, 1.7857142857142858]]
        self.assertEqual(expected, test_case)



from pathlib import Path

from musurgia.pdf.canvas import Canvas
from musurgia.pdf.pdf import Pdf
from musurgia.unittest import TestCase, create_test_path

path = Path(__file__)


class TestCanvas(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    def test_empty_draw(self):
        pdf_path = create_test_path(path, 'empty_draw.pdf')
        canvas = Canvas()
        canvas.draw(self.pdf)
        self.pdf.write(pdf_path)
        self.assertCompareFiles(pdf_path)



import itertools

from musicscore import basic_functions
from musicscore.musictree.midi import MidiNote

from musurgia.random import Random
from musurgia.quantize import get_quantized_positions


class MidiGenerator(object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._node = None

    @property
    def node(self):
        return self._node

    @node.setter
    def node(self, value):
        self._node = value

    def next(self):
        err = str(type(self)) + ' should override next()'
        raise ImportWarning(err)

    def copy(self):
        err = str(type(self)) + ' should override copy()'
        raise ImportWarning(err)


class RelativeMidi(MidiGenerator):
    def __init__(self, midi_range=None, proportions=None, directions=None, microtone=2, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._midi_range = None
        self._proportions = None
        self._directions = None
        self._microtone = None
        self._iterator = None
        self._direction_iterator = None

        self.midi_range = midi_range
        self.proportions = proportions
        self.directions = directions
        self.microtone = microtone

        self._auto_ranged = False

    @property
    def midi_range(self):
        return self._midi_range

    @midi_range.setter
    def midi_range(self, values):
        def get_midi_note_values():
            output = values
            for index, v in enumerate(values):
                if isinstance(v, MidiNote):
                    output[index] = v.value
            return output

        if values:

            try:
                if not hasattr(values, '__iter__'):
                    values = [values, values]
                if len(values) == 1:
                    value = values[0]
                    values = [value, value]

                if len(values) != 2:
                    raise ValueError('wrong length for midi_range')
            except:
                raise TypeError('wrong type for midi_range')

            values = get_midi_note_values()

            if min(values) < 18:
                raise ValueError('midi cannot be smaller than 18')

            self._midi_range = values
        else:
            self._midi_range = None

    @property
    def proportions(self):
        if not self._proportions:
            try:
                self.proportions = self.node.children_fractal_values
            except Exception as err:
                pass
        return self._proportions

    @proportions.setter
    def proportions(self, values):
        if values is not None:
            self._proportions = [value / sum(values) * 100 for value in values]
        else:
            self._proportions = None

    @property
    def directions(self):
        return self._directions

    @directions.setter
    def directions(self, values):
        # print('directions {} setting for {} and node {}'.format(values, self, self.node))
        if values:
            for value in values:
                if value not in [-1, 1]:
                    raise ValueError('directions can only be 1 or -1')

            self._directions = values
            self._direction_iterator = itertools.cycle(self._directions)
        else:
            self._directions = None
            self._direction_iterator = None

    def set_directions(self, *values):
        if len(values) != len(self.proportions):
            raise ValueError(
                "values length {} must equal proportions length {}".format(len(values), len(self.proportions)))
        self.directions = list(values)
        if self.node:
            root = self.node.get_root()
            if root.permute_directions:
                permutation_dict = {}
                for index, order in enumerate(root.permutation_order):
                    permutation_dict[order] = self.directions[index]
                keys = sorted(permutation_dict.keys())
                tree_directions = [permutation_dict[key] for key in keys]
                root.tree_directions = tree_directions
            else:
                root.tree_directions = self.directions

    @property
    def direction_iterator(self):
        return self._direction_iterator

    @property
    def microtone(self):
        return self._microtone

    @microtone.setter
    def microtone(self, value):
        if value and value not in [2, 4, 8]:
            raise ValueError('microtone can only be 2,4,8 or None')
        self._microtone = value

    @property
    def iterator(self):

        if self._iterator is None:
            def scale(old_value, old_lower_limit, old_higher_limit, new_lower_limit, new_higher_limit):
                old_range = old_higher_limit - old_lower_limit
                if old_range == 0:
                    new_value = new_lower_limit
                else:
                    new_range = (new_higher_limit - new_lower_limit)
                    new_value = (((old_value - old_lower_limit) * new_range) / old_range) + new_lower_limit
                return new_value

            if not self.directions:
                raise AttributeError('set directions')
            if not self.proportions:
                raise AttributeError('set proportions')
            if not self.midi_range:
                raise AttributeError('set midi_range')
            if not self.microtone:
                raise AttributeError('set microtone')

            intervals = [proportion * self.direction_iterator.__next__() for proportion in self.proportions]
            midis = basic_functions.dToX(intervals)
            midis = [scale(midi, min(midis), max(midis), min(self.midi_range), max(self.midi_range)) for midi in midis]

            grid = 2 / self.microtone
            quantized_positions = get_quantized_positions(midis, grid_size=grid)
            quantized_midis = [float(midi) for midi in quantized_positions]

            self._iterator = iter(quantized_midis)

        return self._iterator

    def next(self):
        return self.iterator.__next__()

    def copy(self):
        return self.__class__(microtone=self.microtone)

    def __deepcopy__(self, memodict={}):
        return self.__class__(midi_range=self.midi_range, proportions=self.proportions, directions=self.directions,
                              microtone=self.microtone)


class RandomMidi(MidiGenerator):
    def __init__(self, pool=None, periodicity=None, seed=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._my_random = Random()

        self.pool = pool
        self.periodicity = periodicity
        self.seed = seed

    @property
    def my_random(self):
        return self._my_random

    @property
    def seed(self):
        return self.my_random.seed

    @seed.setter
    def seed(self, value):
        self.my_random.seed = value

    @property
    def pool(self):
        return self._my_random.pool

    @pool.setter
    def pool(self, values):
        if values is not None:
            if min(values) < 18:
                raise ValueError('midi cannot be smaller than 18')

            self._my_random.pool = list(set(values))
        else:
            self._my_random.pool = None

    @property
    def periodicity(self):
        return self._my_random.periodicity

    @periodicity.setter
    def periodicity(self, value):
        self._my_random.periodicity = value

    @property
    def iterator(self):
        return self.my_random

    def next(self):
        return self.my_random.__next__()

    def copy(self):
        return self.__class__(pool=self.pool, seed=self.seed, periodicity=self.periodicity)



from unittest import TestCase

from musurgia.fractaltree.fractaltree import FractalTree


class Test(TestCase):
    def setUp(self) -> None:
        ft = FractalTree(proportions=[1, 2, 3], tree_permutation_order=[3, 1, 2], value=20)
        ft.add_layer()
        ft.get_children()[1].add_layer()
        self.ft = ft

    def test_1(self):
        self.assertIsNone(self.ft.get_previous_sibling())

    def test_2(self):
        self.assertIsNone(self.ft.get_children()[0].get_previous_sibling())

    def test_3(self):
        self.assertEqual(self.ft.get_children()[2].get_previous_sibling(), self.ft.get_children()[1])

    def test_4(self):
        self.assertEqual(self.ft.get_children()[1].get_children()[2].get_previous_sibling(),
                         self.ft.get_children()[1].get_children()[1])

    def test_5(self):
        self.assertIsNone(self.ft.get_children()[1].get_children()[0].get_previous_sibling())



from unittest import TestCase

from musurgia.tree import Tree


class Test(TestCase):
    def test_1(self):
        t = Tree()
        self.assertIsNone(t.previous_sibling)

    def test_2(self):
        t = Tree()
        t.add_child(Tree())
        t.add_child(Tree())
        self.assertIsNone(t.get_children()[0].previous_sibling)

    def test_3(self):
        t = Tree()
        t.add_child(Tree())
        t.add_child(Tree())
        self.assertEqual(t.get_children()[1].previous_sibling, t.get_children()[0])



import os
import unittest
from pathlib import Path

from diff_pdf_visually import pdfdiff


def create_test_path(path, test_name):
    return path.parent.joinpath(f'{path.stem}_{test_name}')


class TestFilePath:
    def __init__(self, unittest, parent_path, name, extension):
        self._unittest = None
        self._parent_path = None
        # self._name = None
        # self._extension = None

        self.unittest = unittest
        self.parent_path = parent_path
        self.name = name
        self.extension = extension
        self.out_path = None

    @property
    def unittest(self):
        return self._unittest

    @unittest.setter
    def unittest(self, val):
        if not isinstance(val, TestCase):
            raise TypeError(f"unittest.value must be of type {type(TestCase)} not{type(val)}")
        self._unittest = val

    @property
    def parent_path(self):
        return self._parent_path

    @parent_path.setter
    def parent_path(self, val):
        if not isinstance(val, Path):
            raise TypeError(f"parent_path.value must be of type {type(Path)} not{type(val)}")
        self._parent_path = val

    def __enter__(self):
        self.out_path = create_test_path(self.parent_path, self.name + '.' + self.extension)
        return self.out_path

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unittest.assertCompareFiles(self.out_path)


class TestCase(unittest.TestCase):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _compare_pdfs(self, actual_file_path, expected_file_path, verbosity):
        self.assertTrue(pdfdiff(actual_file_path, expected_file_path, verbosity=verbosity))

    def _compare_contents(self, actual_file_path, expected_file_path):
        with open(actual_file_path, 'r') as myfile:
            result = myfile.read()

        with open(expected_file_path, 'r') as myfile:
            expected = myfile.read()

        self.assertEqual(expected, result)

    def assertCompareFiles(self, actual_file_path, expected_file_path=None, verbosity=0):
        file_name, extension = os.path.splitext(actual_file_path)
        if not expected_file_path:
            if not extension:
                raise ValueError('actual_file_path has no file extension')
            expected_file_path = file_name + '_expected' + extension

        if extension == '.pdf':
            self._compare_pdfs(actual_file_path, expected_file_path, verbosity=verbosity)
        else:
            self._compare_contents(actual_file_path, expected_file_path)

    def file_path(self, parent_path, name, extension):
        tfp = TestFilePath(self, parent_path, name, extension)
        return tfp



import os
from itertools import cycle

from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from quicktions import Fraction

from musurgia.chordfield.chordfield import ChordField
from musurgia.chordfield.valuegenerator import ValueGenerator
from musurgia.fractaltree.fractalmusic import FractalMusic

path = str(os.path.abspath(__file__).split('.')[0])

fm = FractalMusic(proportions=(1, 2, 3, 4, 5, 6, 7), tree_permutation_order=(2, 6, 4, 1, 3, 7, 5), tempo=56,
                  duration=45 / 2, multi=(2, 6), reading_direction='vertical')
fm.generate_children(number_of_children=4)
fm.quantize_children(grid_size=1)
selected_child = [ch for ch in fm.get_children() if ch.fractal_order == 4][0]
selected_child.midi_value = 0

selected_child.chord_field = ChordField(duration_generator=ValueGenerator(cycle([Fraction(1, 5)])),
                                        midi_generator=ValueGenerator(cycle([selected_child.midi_value])))

score = TreeScoreTimewise()
fm.get_simple_format(layer=fm.number_of_layers).to_stream_voice().add_to_score(score)

xml_path = path + '.xml'

score.write(xml_path)



from pathlib import Path

from musurgia.pdf.line import StraightLine
from musurgia.pdf.masterslave import Master
from musurgia.pdf.pdf import Pdf
from musurgia.unittest import TestCase

path = Path(__file__)


class DummyMaster(Master):
    def get_slave_margin(self, slave, margin):
        return 10

    def get_slave_position(self, slave, position):
        return 20


class TestStraightLine(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.master = DummyMaster()
        self.sl = StraightLine(mode='h', length=20, name='straight_test', master=self.master)

    def test_relative_x(self):
        actual = self.sl.relative_x
        expected = self.master.get_slave_position(self.sl, 'x')
        self.assertEqual(expected, actual)

    def test_get_relative_x(self):
        actual = self.sl.get_relative_x2()
        expected = self.sl.relative_x + self.sl.length
        self.assertEqual(expected, actual)

    def test_get_width(self):
        actual = self.sl.get_width()
        expected = self.sl.left_margin + self.sl.length + self.sl.right_margin
        self.assertEqual(expected, actual)

    def test_get_height(self):
        actual = self.sl.get_height()
        expected = self.sl.top_margin + self.sl.bottom_margin
        self.assertEqual(expected, actual)

    def test_draw(self):
        with self.file_path(path, 'draw', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.sl.draw(self.pdf)
            self.pdf.write(pdf_path)



from unittest import TestCase

from musurgia.fractaltree.fractalmusic import FractalMusic, TempoIsAlreadySet, ChildTempoIsAlreadySet


class Test(TestCase):
    def test_1(self):
        fm = FractalMusic(duration=100)
        fm.tempo = 72
        fm.add_layer()
        with self.assertRaises(TempoIsAlreadySet):
            fm.get_children()[0].tempo = 60

    def test_2(self):
        fm = FractalMusic(duration=100)
        fm.add_layer()
        fm.get_children()[0].tempo = 72
        with self.assertRaises(ChildTempoIsAlreadySet):
            fm.tempo = 60

    def test_3(self):
        fm = FractalMusic(duration=100)
        fm.add_layer()
        try:
            fm.tempo = 60
        except (ChildTempoIsAlreadySet, TempoIsAlreadySet) as err:
            self.fail(err + "was raised!")

    def test_4(self):
        fm = FractalMusic(duration=100)
        fm.add_layer()
        fm.get_children()[0].tempo = 72
        fm.set_none_tempi(90)
        self.assertEqual([child.tempo for child in fm.get_children()], [72, 90., 90])

    def test_5(self):
        fm = FractalMusic(duration=100)
        fm.add_layer()
        fm.tempo = 72
        self.assertEqual([child.tempo for child in fm.get_children()], [72, 72, 72])

    def test_6(self):
        fm = FractalMusic(duration=100)
        fm.add_layer()

        fm.get_children()[1].tempo = 80
        fm.add_layer()
        fm.get_children()[2].get_children()[2].tempo = 72
        fm.add_layer()

        self.assertEqual(fm.get_layer(2, key='tempo'), [[None, None, None], [80, 80, 80], [None, None, 72]])

    def test_7(self):
        fm = FractalMusic(duration=100)
        fm.add_layer()

        fm.get_children()[1].tempo = 80
        fm.add_layer()
        fm.get_children()[2].get_children()[2].tempo = 72
        fm.add_layer()
        fm.set_none_tempi(60)

        self.assertEqual(fm.get_layer(2, key='tempo'), [[60, 60, 60], [80, 80, 80], [60, 60, 72]])

    def test_8(self):
        fm = FractalMusic(duration=100)
        fm.tempo = 72
        fm.add_layer()
        fm.add_layer()
        fm.add_layer()
        fm.set_none_tempi(60)

        self.assertEqual(fm.get_layer(2, key='tempo'), [[72, 72, 72], [72, 72, 72], [72, 72, 72]])



from musurgia.pdf.drawobject import DrawObject
from musurgia.pdf.labeled import Labeled
from musurgia.pdf.segmentedline import MarkLine, LineSegment
from musurgia.pdf.named import Named


class Cell(DrawObject):
    def __init__(self, width=10, height=10, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._width = None
        self._height = None
        self.width = width
        self.height = height
        self._lines = None

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, val):
        self._width = val
        self.generate_lines()

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, val):
        self._height = val
        self.generate_lines()

    @DrawObject.relative_y.setter
    def relative_y(self, val):
        super().relative_y = val
        self.generate_lines()

    @DrawObject.relative_x.setter
    def relative_x(self, val):
        super().relative_x = val
        self.generate_lines()

    def generate_lines(self):
        if self.width and self.height:
            # todo: LineSegment must have an angel property
            line_1 = LineSegment(relative_y=self.relative_y, relative_x=self.relative_x)


class Table(DrawObject):
    def __init__(self, number_of_rows=10, number_of_columns=10, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._number_of_rows = None
        self._number_of_columns = None

        self.number_of_rows = number_of_rows
        self.number_of_columns = number_of_columns

        self._page = None

    @property
    def number_of_rows(self):
        return self._number_of_rows

    @number_of_rows.setter
    def number_of_rows(self, val):
        self._number_of_rows = val

    @property
    def number_of_columns(self):
        return self._number_of_columns

    @number_of_columns.setter
    def number_of_columns(self, val):
        self._number_of_columns = val

    def get_row(self, row_number):
        pass

    def get_column(self, column_number):
        pass

    def get_cell(self, row_number, column_number):
        pass

    def draw(self, pdf):
        pass
        # x1 = pdf.x + self.relative_x
        # x2 = pdf.x + self.relative_x + self.actual_length
        # y1 = pdf.y + self.relative_y
        # y2 = y1
        #
        # self._x1, self._x2, self._y1, self._y2 = x1, x2, y1, y2
        # self._page = pdf.page
        # if self.name:
        #     self.name.draw(pdf)
        #
        # if self.show:
        #     self.start_mark_line.draw(pdf)
        #     self.end_mark_line.draw(pdf)
        #     pdf.line(x1=x1, y1=y1, x2=x2, y2=y2)
        #     for text_label in self._text_labels:
        #         text_label.draw(pdf)
        #
        # pdf.x = x2



from pathlib import Path

from quicktions import Fraction

from musurgia.pdf.pdf import Pdf
from musurgia.timeline.timeline import TimeLine
from musurgia.unittest import TestCase, create_test_path

path = Path(__file__)


class Test(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    # def test_ruler_bottom_margin(self):
    #     tl = TimeLine(length=30, bottom_margin=40)
    #     tl.add_voice('v1')
    #     print(tl.bottom_margin)
    #     print(tl.ruler.bottom_margin)
    #     print(tl.get_height())

    def test_draw_ruler(self):
        pdf_path = create_test_path(path, 'draw_ruler.pdf')
        pdf = Pdf(orientation='p')
        tl = TimeLine(length=50, bottom_margin=40)
        printable = (pdf.w - pdf.r_margin - pdf.l_margin)
        tl.unit = Fraction(Fraction(printable), Fraction(20))
        tl.show_interval = 10
        tl.draw(pdf)
        pdf.write(pdf_path)
        self.assertCompareFiles(actual_file_path=pdf_path)

    def test_draw_ruler_with_break(self):
        pdf_path = create_test_path(path, 'draw_ruler_with_break.pdf')
        pdf = Pdf(orientation='p')
        tl = TimeLine(length=300)
        printable = (pdf.w - pdf.r_margin - pdf.l_margin)
        tl.unit = Fraction(Fraction(printable), Fraction(20))
        tl.show_interval = 10
        tl.draw(pdf)
        pdf.write(pdf_path)
        self.assertCompareFiles(actual_file_path=pdf_path)

    def test_time_line_with_voices(self):
        pdf_path = create_test_path(path, 'time_line_with_voices.pdf')
        tl = TimeLine(length=200, inner_distance=20, bottom_margin=30)
        # voice = tl.add_voice(name='v1')
        # voice.name.relative_x = -5
        # for segment in voice.segments[5:15]:
        #     segment.show = True

        tl.draw(self.pdf)
        self.pdf.write(pdf_path)



from musicscore.musictree.treechord import TreeChord

from musurgia.fractaltree.fractalmusic import FractalMusic

fm = FractalMusic(tempo=60, quarter_duration=10)
fm.add_layer()
fm.get_children()[1].chord.add_grace_chords(chords=[TreeChord(midis=[63, 65]), TreeChord(midis=[64]), TreeChord(midis=[66])])



from musurgia.pdf.text import TextLabel
from musurgia.timeline.abstractvoice import AbstractVoice
from musurgia.timing import Timing


class Ruler(AbstractVoice):
    def __init__(self, show_interval=5, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._show_interval = None
        self.show_interval = show_interval

    @property
    def show_interval(self):
        return self._show_interval

    @show_interval.setter
    def show_interval(self, val):
        if not isinstance(val, int):
            raise TypeError('show_interval.value must be of type int not{}'.format(type(val)))
        self._show_interval = val
        self.update_show_intervals()

    def update_show_intervals(self):
        for index, line_segment in enumerate(self.line_segments):
            seconds = index
            line_segment.start_mark_line.remove_text_labels()
            if seconds % self.show_interval == 0:
                line_segment.start_mark_line.add_text_label(
                    TextLabel(text=Timing.get_clock(seconds, mode='ms'), relative_y=-3, font_size=9))
                line_segment.start_mark_line.thickness = 4



import os

from musicscore.musictree.midi import C

from musurgia.unittest import TestCase
from musurgia.fractaltree.fractalmusic import FractalMusic

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.fm = FractalMusic(proportions=[1, 2, 3, 4], tree_permutation_order=[3, 1, 4, 2], quarter_duration=20,
                               tempo=70)
        self.fm.midi_generator.midi_range = [36, 60]

    def test_1(self):
        self.fm.add_layer()
        score = self.fm.get_score(layer_number=1)

        for leaf in self.fm.traverse_leaves():
            leaf.set_chord(leaf.get_choral(range_factor=1.5, direction=1, last=True))

        sf = self.fm.get_simple_format()

        v = sf.to_stream_voice(1)
        v.add_to_score(score, 2)

        xml_path = path + '_test_1.xml'
        score.write(path=xml_path)

        self.assertCompareFiles(actual_file_path=xml_path)

    def test_2(self):
        self.fm.midi_generator.midi_range = [C(4).value, C(6).value]
        self.fm.midi_generator.directions = [1, -1, 1]
        self.fm.add_layer()
        score = self.fm.get_score(layer_number=1)

        for leaf in self.fm.traverse_leaves():
            leaf.set_chord(leaf.get_choral(range_factor=1, direction=1, last=False))
        sf = self.fm.get_simple_format()

        v = sf.to_stream_voice(1)
        v.add_to_score(score, 2)

        for leaf in self.fm.traverse_leaves():
            leaf.add_layer()

        sf = self.fm.get_simple_format()

        v = sf.to_stream_voice(1)
        v.add_to_score(score, 3)

        xml_path = path + '_test_2.xml'
        score.write(path=xml_path)

        self.assertCompareFiles(actual_file_path=xml_path)

    def test_3(self):
        fm = FractalMusic(proportions=(1, 2, 3, 4, 5, 6, 7), tree_permutation_order=(2, 6, 4, 1, 3, 7, 5),
                          quarter_duration=30, tempo=80)
        fm.midi_generator.midi_range = (60, 72)
        fm.midi_generator.microtone = 4
        fm.add_layer()
        fm.quantize_children(grid_size=1)
        expected = [child.get_choral_midis() for child in fm.get_children()]
        actual = [child.get_choral_midis() for child in fm.get_children()]
        self.assertEqual(expected, actual)



from unittest import TestCase

from musurgia.fractaltree.fractalmusic import FractalMusic
import os

path = os.path.abspath(__file__).split(".")[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.fm = FractalMusic(proportions=[1, 2, 3], tree_permutation_order=(3, 1, 2), duration=10)

    def test_1(self):
        self.fm.midi_generator.set_directions(1, -1, -1)
        self.assertEqual([1, -1, -1], self.fm.midi_generator.directions)
        self.assertEqual([-1, -1, 1], self.fm.tree_directions)
        self.fm.add_layer()
        # for node in self.fm.get_children():
        #     print(node.name)
        #     print(node.multi)
        #     print(node.permutation_order)
        #
        # text_path = path + '_test_1.txt'
        #
        # self.fm.write_infos(text_path)
        directions = [leaf.midi_generator.directions for leaf in self.fm.traverse_leaves()]

        self.assertEqual([[-1, -1, 1], [1, -1, -1], [-1, 1, -1]], directions)

    def test_2(self):
        self.fm.tree_directions = [1, 1, -1]
        self.fm.midi_generator.midi_range = [60, 70]
        self.fm.add_layer()
        directions = [leaf.midi_generator.directions for leaf in self.fm.traverse_leaves()]
        self.fm.add_layer()
        midis = [leaf.midi_value for leaf in self.fm.traverse_leaves()]
        self.assertEqual([[1, 1, -1], [-1, 1, 1], [1, -1, 1]], directions)
        self.assertEqual([60.0, 63.0, 70.0, 63.0, 60.0, 61.0, 65.0, 70.0, 63.0], midis)

    def test_3(self):
        self.fm.tree_directions = [1, 1, -1]
        self.fm.midi_generator.midi_range = [60, 70]
        self.fm.permute_directions = False
        self.fm.add_layer()
        self.fm.add_layer()
        midi_ranges = [[leaf.midi_generator.midi_range for leaf in child.get_leaves()] for child in
                       self.fm.get_children()]
        result = [[[60.0, 63.0], [63.0, 70.0], [70.0, 60.0]], [[60.0, 62.0], [62.0, 63.0], [63.0, 61.0]],
                  [[63.0, 66.0], [66.0, 70.0], [70.0, 69.0]]]

        self.assertEqual(result, midi_ranges)



from fractions import Fraction
from unittest import TestCase

from musurgia.fractaltree.fractaltree import FractalTree


class Test(TestCase):
    def setUp(self) -> None:
        self.ft = FractalTree(proportions=[1, 2, 3], tree_permutation_order=[3, 1, 2], value=10)

    def test_0(self):
        with self.assertRaises(Exception):
            self.ft.get_layer(1)

    def test_1(self):
        self.assertEqual([self.ft], self.ft.get_layer(0))

    def test_2(self):
        self.ft.add_layer()
        result = self.ft.get_children()
        self.assertEqual(result, self.ft.get_layer(1))

    def test_3(self):
        for i in range(3):
            self.ft.add_layer()

        result = self.ft.get_children()
        self.assertEqual(result, self.ft.get_layer(1))

    def test_4(self):
        for i in range(3):
            self.ft.add_layer()
        result = [child.get_children() for child in self.ft.get_children()]

        self.assertEqual(result, self.ft.get_layer(2))

    def test_5(self):
        for i in range(3):
            self.ft.add_layer()
        result = self.ft.get_leaves()

        self.assertEqual(result, self.ft.get_layer(3))

    def test_6(self):
        for i in range(3):
            self.ft.add_layer()

        with self.assertRaises(ValueError):
            self.ft.get_layer(4)

    def test_7(self):
        self.ft.add_layer()
        self.ft.add_layer(lambda n: True if n.fractal_order > 1 else False)
        self.ft.add_layer(lambda n: True if n.fractal_order > 1 else False)
        self.ft.add_layer(lambda n: True if n.fractal_order > 1 else False)
        self.ft.add_layer(lambda n: True if n.fractal_order > 1 else False)
        result = [[['1.1'], [['1.2.1'], ['1.2.2.1', '1.2.2.2', '1.2.2.3'], ['1.2.3.1', '1.2.3.2', '1.2.3.3']],
                   [['1.3.1.1', '1.3.1.2', '1.3.1.3'], ['1.3.2'], ['1.3.3.1', '1.3.3.2', '1.3.3.3']]], '2',
                  [[['3.1.1'], ['3.1.2.1', '3.1.2.2', '3.1.2.3'], ['3.1.3.1', '3.1.3.2', '3.1.3.3']],
                   [['3.2.1.1', '3.2.1.2', '3.2.1.3'], ['3.2.2'], ['3.2.3.1', '3.2.3.2', '3.2.3.3']], ['3.3']]]
        self.assertEqual(result, [name for name in self.ft.get_layer(4, key='name')])

    def test_7_1(self):
        self.ft.add_layer()
        self.ft.add_layer()
        self.assertEqual([10], self.ft.get_layer(0, key='value'))

    def test_7_2(self):
        self.ft.add_layer()
        self.ft.add_layer()
        result = [Fraction(5, 1), Fraction(5, 3), Fraction(10, 3)]
        self.assertEqual(result, self.ft.get_layer(1, key='value'))

    def test_7_3(self):
        self.ft.add_layer()
        self.ft.add_layer()
        result = [[Fraction(5, 6), Fraction(5, 3), Fraction(5, 2)], [Fraction(5, 6), Fraction(5, 18), Fraction(5, 9)],
                  [Fraction(10, 9), Fraction(5, 3), Fraction(5, 9)]]
        self.assertEqual(result, self.ft.get_layer(2, key='value'))

    def test_get_layer_key_lambda(self):
        self.ft.add_layer()
        self.ft.add_layer()
        result = [[0.83, 1.67, 2.5], [0.83, 0.28, 0.56], [1.11, 1.67, 0.56]]
        self.assertEqual(result, self.ft.get_layer(2, key=lambda node: round(float(node.value), 2)))



import os

from musicscore.basic_functions import xToD
from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

from musurgia.unittest import TestCase
from musurgia.gearwheels import GearWheels, Wheel

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def test_1(self):
        wh = Wheel(size=4, start=60)
        expected = [60, 64, 68, 72, 76]
        actual = [position.value for position in wh.get_positions(number_of_cycles=5)]
        self.assertEqual(expected, actual)

    def test_2(self):
        gw = GearWheels(wheels=[Wheel(3), Wheel(4), Wheel(5)])
        expected = [3, 1, 1, 1, 2, 1, 1, 2, 3, 1, 2, 2, 1, 3, 1, 2, 1, 2, 2, 1, 2, 1, 3, 1, 2, 2, 1, 3, 2, 1, 1, 2, 1,
                    1, 1, 3]
        actual = gw.get_rhythm()
        self.assertEqual(actual, expected)

    def test_3(self):
        xml_path = path + '_test_3.xml'
        score = TreeScoreTimewise()
        gw = GearWheels(wheels=[Wheel(3), Wheel(4), Wheel(5)])
        sfs = []
        for index, wheel in enumerate(gw.wheels):
            quarter_durations = xToD(wheel.get_position_values())
            midis = len(quarter_durations) * [60 + index]
            sf = SimpleFormat(quarter_durations=quarter_durations, midis=midis)
            sfs.append(sf)
            sf.to_stream_voice().add_to_score(score, part_number=index + 1)
        sf = SimpleFormat.sum(*sfs)
        sf.to_stream_voice().add_to_score(score, part_number=len(gw.wheels) + 1)
        score.write(xml_path)
        self.assertCompareFiles(xml_path)



from musurgia.fractaltree.fractalmusic import FractalMusic
from musurgia.unittest import TestCase


class Test(TestCase):
    def test_get_durational_position_in_tree(self):
        fm = FractalMusic(duration=10)
        fm.add_layer()
        actual = fm.get_children()[2].durational_position_in_tree
        expected = sum(ch.duration for ch in fm.get_children()[:2])
        self.assertEqual(expected, actual)
